import networkx as nx
import numpy as np
import tensorflow as tf
import scipy as sp
import matplotlib.pyplot as plt
import h5py

import os, random, sys
import time

from tensorflow.keras import Model, Input
from tensorflow.keras.layers import Conv1D, Dense

mname = sys.argv[1]

VSEED = int(sys.argv[2])

#setting random seed, for reproducibility
os.environ['PYTHONHASHSEED']=str(VSEED)
random.seed(VSEED)
np.random.seed(VSEED)
tf.random.set_seed(VSEED)

gpu = tf.config.experimental.list_physical_devices('GPU')
for gp in gpu:
    tf.config.experimental.set_memory_growth(gp, True)

tf.keras.backend.set_floatx('float64')
tf.compat.v1.disable_eager_execution()

BATCH_SIZE = 8

    
DXes = []
DYes = []   
data_dir ="../processed_data"
for fname in os.listdir(data_dir):
    with h5py.File(data_dir+"/"+fname, 'r') as f:
        dd = f['data'][:, :, np.array([3,4,5,12,13,14])].astype('float64')
        DXes.append(
            np.concatenate(
                [
                dd,
                np.tile(f['args'][:],(dd.shape[0],dd.shape[1],1))
                ],
                axis=2
            )
            )
        DYes.append(f['data'][:, :, -1:].astype('float64'))

allDX = np.concatenate(DXes, axis=0)
allDY = np.log(np.concatenate(DYes, axis=0))

print(allDX.shape)


allidxs = np.arange(allDX.shape[0])
np.random.shuffle(allidxs)

TT_split = int(.8*allidxs.shape[0])

print(allidxs[:2])
#train_idxs = np.concatenate(
#    [allidxs[:2] for i in range(10*BATCH_SIZE)]
#    )
#test_idxs = train_idxs[:2*BATCH_SIZE]    

train_idxs = allidxs[:TT_split]    
test_idxs = allidxs[TT_split:]

meanX = np.mean(allDX[train_idxs,:,:],axis=0,keepdims=True)
stdX = .00000001+np.std(allDX[train_idxs,:,:],axis=0,keepdims=True)

meanY = np.mean(allDY[train_idxs,:,:],axis=0,keepdims=True)
stdY = .00000001+np.std(allDY[train_idxs,:,:],axis=0,keepdims=True)


normX = (allDX - meanX)/stdX
normY =  (allDY - meanY)/stdY


def make_data_set(idxs):
    the_dataset = tf.data.Dataset.from_tensor_slices((normX[idxs], normY[idxs]))
    return the_dataset

train_ds = make_data_set(train_idxs).shuffle(50000).batch(BATCH_SIZE,drop_remainder=True).repeat().prefetch(tf.data.experimental.AUTOTUNE)
valid_ds = make_data_set(test_idxs).batch(BATCH_SIZE,drop_remainder=True).prefetch(tf.data.experimental.AUTOTUNE)

from networkx_builder import *

mname = sys.argv[1]
diffpool = "diff" in sys.argv[1]
adaptive = "adaptive" in sys.argv[1]
THE_INPUT = Input(batch_shape=(BATCH_SIZE,624,10))
diffpool_flopcost = 0.0

if not diffpool:

    l1 = np.loadtxt('../l1.csv',delimiter=",").astype('float64')
    l2 = np.loadtxt('../l2.csv',delimiter=",").astype('float64')
    l3 = np.loadtxt('../l3.csv',delimiter=",").astype('float64')

    p32 = np.loadtxt('../p32.csv',delimiter=",").astype('float64')
    p31 = np.dot(
            np.loadtxt('../p32.csv',delimiter=","),
            np.loadtxt('../p21.csv',delimiter=",")
          ).astype('float64')
    p21 = np.loadtxt('../p21.csv',delimiter=",").astype('float64')
    
    p32V = tf.Variable(initial_value=p32,
                         trainable=adaptive,name='p32')
                         
    p21V = tf.Variable(initial_value=p21,
                         trainable=adaptive,name='p21')
else:
    
    l3 = np.loadtxt('../l3.csv',delimiter=",").astype('float64')
    diffMGCN = GCNModel(l3, [16,32,64],act=tf.nn.relu)
    diffMFinalLayer = GraphConvLayer(l3,ft_out=312,act=tf.nn.softmax)
    p32V = diffMFinalLayer(
            diffMGCN(
                tf.transpose(THE_INPUT, perm=[1,0,2])
            )
        )
    p32V = tf.reduce_mean(p32V,axis=1)
        
    l2 = tf.tensordot(
        tf.tensordot(
            p32V,
            l3,
            axes=[[0],[0]]
        ),
        p32V,
        axes=[[-1],[0]]
    )
    
    ccc = tf.tensordot(
              p32V,
              THE_INPUT,
              axes=[[0],[1]]
            )
    #ccc = tf.transpose(
    #    ccc,
    #    perm=[1,0,2]
    #)
    #print(ccc.shape)
    #quit()    
    diffCGCN = GCNModel(l2, [16,32,64],act=tf.nn.relu)
    diffCFinalLayer = GraphConvLayer(l2,ft_out=72,act=tf.nn.softmax)
    p21V = diffCFinalLayer(
            diffCGCN(ccc)
        )
    p21V = tf.reduce_mean(p21V, axis=1)    
    l1 = tf.tensordot(
        tf.tensordot(
            p21V,
            l2,
            axes=[[0],[0]]
        ),
        p21V,
        axes=[[-1],[0]]
    )              
    diffpool_flopcost += (diffCGCN.flopcost + diffCFinalLayer.flopcost + diffMGCN.flopcost + diffMFinalLayer.flopcost)
    #diffpool_flopcost += 312*624*13 + 312*72*13 + 624*624*312 + 312*312*72


fine_model = build_nested_model([l3, l2, l1],[p32V, p21V])[0]

fine_model.build(input_shape=(BATCH_SIZE, 624, 10))


model = Model([THE_INPUT],[fine_model(THE_INPUT)],name=mname)
model.flopcost = (fine_model.flopcost + diffpool_flopcost)/20.0

model.build(input_shape=(BATCH_SIZE, 624, 10))

model.compile(optimizer=tf.keras.optimizers.Adam(),
              loss=tf.keras.losses.MeanSquaredError(),
              metrics=['mse'])

model.summary()

INITIAL_TIME = time.time()
TOTAL_COST = 1.0

class CostCallback(tf.keras.callbacks.Callback):
    def __init__(self):
        super(CostCallback, self).__init__()
    
        
    def on_train_batch_end(self,batch,logs=None):
        global TOTAL_COST
        TOTAL_COST += self.model.flopcost

class LoggingAndSavingCallback(tf.keras.callbacks.Callback):
    def __init__(self):
        super(LoggingAndSavingCallback, self).__init__()
        self.best_weights=None
        self.best_validation=-1 
        
    def on_epoch_end(self, epoch, logs=None):
        global TOTAL_COST
        if 'val_loss' in logs.keys():
            val = logs['val_loss']
            print(self.model.name, (epoch+1)*self.model.flopcost, val, time.time()-INITIAL_TIME)
            if self.best_validation < 0 or val < self.best_validation:
                self.best_validation = val           
                self.model.save_weights("saved_models/" + self.model.name + "_" + str(VSEED) +"/best_ver", save_format='tf')  

val1 = model.evaluate(valid_ds,verbose=0)
print(model.name, 1.0, val1[0], 0.0)

model.fit(train_ds,
          validation_data=valid_ds,
          validation_freq=1,
          epochs=2000,
          steps_per_epoch=20,
          verbose=0,
          callbacks=[LoggingAndSavingCallback(),CostCallback()]
          )
          
quit()          


