import networkx as nx
import numpy as np
import tensorflow as tf
import scipy as sp
import matplotlib.pyplot as plt
import h5py

import os, random, sys

from time import time

from tensorflow.keras import Model
from tensorflow.keras.layers import Conv1D, Dense

mname = sys.argv[1]
RSEED = int(sys.argv[2])

#setting random seed, for reproducibility
os.environ['PYTHONHASHSEED']=str(RSEED)
random.seed(RSEED)
np.random.seed(RSEED)
tf.random.set_seed(RSEED)

gpu = tf.config.experimental.list_physical_devices('GPU')
for gp in gpu:
    tf.config.experimental.set_memory_growth(gp, True)

tf.keras.backend.set_floatx('float64')


BATCH_SIZE = 8

    
DXes = []
DYes = []   
data_dir ="../processed_data"
for fname in os.listdir(data_dir):
    with h5py.File(data_dir+"/"+fname, 'r') as f:
        dd = f['data'][:, :, np.array([3,4,5,12,13,14])].astype('float64')
        DXes.append(
            np.concatenate(
                [
                dd,
                np.tile(f['args'][:],(dd.shape[0],dd.shape[1],1))
                ],
                axis=2
            )
            )
        DYes.append(f['data'][:, :, -1:].astype('float64'))

allDX = np.concatenate(DXes, axis=0)
allDY = np.log(np.concatenate(DYes, axis=0))

print(allDX.shape)


allidxs = np.arange(allDX.shape[0])
np.random.shuffle(allidxs)

TT_split = int(.8*allidxs.shape[0])

print(allidxs[:2])
#train_idxs = np.concatenate(
#    [allidxs[:2] for i in range(10*BATCH_SIZE)]
#    )
#test_idxs = train_idxs[:2*BATCH_SIZE]    

train_idxs = allidxs[:TT_split]    
test_idxs = allidxs[TT_split:]

meanX = np.mean(allDX[train_idxs,:,:],axis=0,keepdims=True)
stdX = .00000001+np.std(allDX[train_idxs,:,:],axis=0,keepdims=True)

meanY = np.mean(allDY[train_idxs,:,:],axis=0,keepdims=True)
stdY = .00000001+np.std(allDY[train_idxs,:,:],axis=0,keepdims=True)


normX = (allDX - meanX)/stdX
normY =  (allDY - meanY)/stdY


def make_data_set(idxs):
    the_dataX = tf.data.Dataset.from_tensor_slices(normX[idxs])
    the_dataY = tf.data.Dataset.from_tensor_slices(normY[idxs])
    the_dataset = tf.data.Dataset.zip((the_dataX, the_dataY))
    return the_dataset

train_ds = make_data_set(train_idxs).shuffle(50000).batch(BATCH_SIZE).repeat().prefetch(tf.data.experimental.AUTOTUNE)
valid_ds = make_data_set(test_idxs).batch(BATCH_SIZE).prefetch(tf.data.experimental.AUTOTUNE)

from networkx_builder_v2 import build

model = build(mname)

model.build(input_shape=(BATCH_SIZE, 624, 10))

model.compile(optimizer=tf.keras.optimizers.Adam(),
              loss=tf.keras.losses.MeanSquaredError(),
              metrics=['mse'])
model.summary()

INITIAL_TIME = time()

class LoggingAndSavingCallback(tf.keras.callbacks.Callback):
    def __init__(self):
        super(LoggingAndSavingCallback, self).__init__()
        self.best_weights=None
        self.best_validation=-1
        
    def on_epoch_end(self, epoch, logs=None):
        if 'val_loss' in logs.keys():
            val = logs['val_loss']
            print(self.model.name, (epoch+1)*self.model.flopcost, val, time()-INITIAL_TIME)
            if self.best_validation < 0 or val < self.best_validation:
                self.best_validation = val           
                self.model.save_weights("saved_models/" + self.model.name + "_" + str(RSEED) +"/best_ver", save_format='tf')   

val1 = model.evaluate(normX[test_idxs],normY[test_idxs],batch_size=BATCH_SIZE,verbose=0)
print(model.name, 1.0, val1[0], 0.0)

model.fit(train_ds,
          validation_data=valid_ds,
          validation_freq=10,
          epochs=1000,
          steps_per_epoch=20,
          verbose=0,
          callbacks=[LoggingAndSavingCallback()]
          )
          
quit()          

@tf.function
def train_step(images, labels,i=0):
    with tf.GradientTape() as tape:
        predictions = model(images)
        loss = loss_object(labels, predictions)
        un_loss = loss_object((labels*stdY)+meanY, (predictions*stdY) + meanY) * 6.241
    gradients = tape.gradient(loss, model.trainable_variables)
    optimizer.apply_gradients(zip(gradients, model.trainable_variables))

    train_loss(loss)
    train_un_loss(un_loss)

loss_object = tf.keras.losses.MeanSquaredError()
train_loss = tf.keras.metrics.Mean(name='train_loss')
train_un_loss = tf.keras.metrics.Mean(name='train_un_loss')

optimizer = tf.keras.optimizers.Adam()

EPOCHS = 20

for epoch in range(EPOCHS):
    for i, (images, labels) in enumerate(train_ds):
        train_step(images, labels)

    #for test_images, test_labels in test_ds:
    #    test_step(test_images, test_labels)

    template = 'Epoch {}, Loss: {}, Un-Normalized Loss: {} x meV'
    print(template.format(epoch+1,
                        train_loss.result(), train_un_loss.result()))

    # Reset the metrics for the next epoch
    train_loss.reset_states()
    train_un_loss.reset_states()




