import networkx as nx
import numpy as np
import tensorflow as tf
import scipy as sp
import matplotlib.pyplot as plt
import h5py

import os, random, sys

from time import time

from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Conv1D, Dense
from tensorflow.keras.backend import get_value

from networkx_builder import *

#setting random seed, for reproducibility
os.environ['PYTHONHASHSEED']=str(42)
random.seed(42)
np.random.seed(42)
tf.random.set_seed(42)

gpu = tf.config.experimental.list_physical_devices('GPU')
for gp in gpu:
    tf.config.experimental.set_memory_growth(gp, True)

tf.keras.backend.set_floatx('float64')


BATCH_SIZE = 8
p32 = np.loadtxt('p32.csv',delimiter=",")
p21 = np.loadtxt('p21.csv',delimiter=",")
    
DXes = []
DYes = []   
data_dir ="../processed_data"
for fname in os.listdir(data_dir):
    with h5py.File(data_dir+"/"+fname, 'r') as f:
        dd = f['data'][:, :, np.array([3,4,5,12,13,14])].astype('float64')
        DXes.append(
            np.concatenate(
                [
                dd,
                np.tile(f['args'][:],(dd.shape[0],dd.shape[1],1))
                ],
                axis=2
            )
            )
        DYes.append(f['data'][:, :, -1:].astype('float64'))

allDX = np.concatenate(DXes, axis=0)
allDY = np.log(np.concatenate(DYes, axis=0))

print(allDX.shape)
#print(np.tensordot(allDX, p32, axes=[[1],[0]]).shape)

allidxs = np.arange(allDX.shape[0])
np.random.shuffle(allidxs)

TT_split = int(.8*allidxs.shape[0])

print(allidxs[:2])
#train_idxs = np.concatenate(
#    [allidxs[:2] for i in range(10*BATCH_SIZE)]
#    )
#test_idxs = train_idxs[:2*BATCH_SIZE]    

train_idxs = allidxs[:TT_split]    
test_idxs = allidxs[TT_split:]

meanX = np.mean(allDX[train_idxs,:,:],axis=0,keepdims=True)
stdX = .00000001+np.std(allDX[train_idxs,:,:],axis=0,keepdims=True)

meanY = np.mean(allDY[train_idxs,:,:],axis=0,keepdims=True)
stdY = .00000001+np.std(allDY[train_idxs,:,:],axis=0,keepdims=True)

p32 = np.loadtxt('p32.csv',delimiter=",").astype('float64')
p31 = np.dot(
        np.loadtxt('p32.csv',delimiter=","),
        np.loadtxt('p21.csv',delimiter=",")
      ).astype('float64')
p21 = np.loadtxt('p21.csv',delimiter=",").astype('float64')

normX = (allDX - meanX)/stdX
normY =  (allDY - meanY)/stdY

medNormX = np.swapaxes(np.tensordot(normX, p32, axes=[[1],[0]]),1,2)
medNormY = np.swapaxes(np.tensordot(normY, p32, axes=[[1],[0]]),1,2)

coarseNormX = np.swapaxes(np.tensordot(medNormX, p21, axes=[[1],[0]]),1,2)
coarseNormY = np.swapaxes(np.tensordot(medNormY, p21, axes=[[1],[0]]),1,2)

CYCLE = sys.argv[1]
ada = "adaptive" in CYCLE
CYCLE = CYCLE.split("_")[0]

p32V = tf.Variable(initial_value=p32,
                     trainable=(ada))
                     
p21V = tf.Variable(initial_value=p21,
                     trainable=(ada))

(fine_model, med_model, coarse_model) = build_nested_model(p21V, p32V, adapt=ada)


fine_inp=tf.keras.Input(shape=(624,10))
fine_prime = Model(fine_inp, fine_model(fine_inp), name='fine'+CYCLE)
fine_prime.flopcost = fine_model.flopcost

med_perm = GCNTransposeLayer()
med_to_fine = ProlongLayer(p32V)
med_inp=tf.keras.Input(shape=(312,10))
med_prime = Model(med_inp, med_perm(med_to_fine(med_perm(med_model(med_inp)))), name='med'+CYCLE)
med_prime.flopcost = med_model.flopcost

coarse_perm = GCNTransposeLayer()
coarse_to_med = ProlongLayer(p21V)
coarse_inp=tf.keras.Input(shape=(72,10))
coarse_prime = Model(coarse_inp, coarse_perm(med_to_fine(coarse_to_med(coarse_perm(coarse_model(coarse_inp))))), name='coarse'+CYCLE)
coarse_prime.flopcost = coarse_model.flopcost


def make_data_set(dX,dY,idxs):
    the_dataset = tf.data.Dataset.from_tensor_slices((dX[idxs], dY[idxs]))
    return the_dataset

train_FINE = make_data_set(normX, normY, train_idxs).shuffle(50000).batch(BATCH_SIZE).repeat().prefetch(tf.data.experimental.AUTOTUNE)
valid_FINE = make_data_set(normX, normY, test_idxs).batch(BATCH_SIZE).prefetch(tf.data.experimental.AUTOTUNE)

train_MED = make_data_set(medNormX, normY, train_idxs).shuffle(50000).batch(BATCH_SIZE).repeat().prefetch(tf.data.experimental.AUTOTUNE)
valid_MED = make_data_set(medNormX, normY, test_idxs).batch(BATCH_SIZE).prefetch(tf.data.experimental.AUTOTUNE)

train_COARSE = make_data_set(coarseNormX, normY, train_idxs).shuffle(50000).batch(BATCH_SIZE).repeat().prefetch(tf.data.experimental.AUTOTUNE)
valid_COARSE = make_data_set(coarseNormX, normY, test_idxs).batch(BATCH_SIZE).prefetch(tf.data.experimental.AUTOTUNE)

optt = tf.keras.optimizers.Adam()
lossed = tf.keras.losses.MeanSquaredError()

#all_vars =  fine_model.weights
#p32V = [item for item in all_vars if 'p32' in item.name][0]
#p21V = [item for item in all_vars if 'p21' in item.name][0]

for model in [coarse_prime, med_prime, fine_prime]:
    model.compile(optimizer=optt,
                  loss=lossed,
                  metrics=['mse'])
    model.flopcost *= 1/20.0

INITIAL_TIME = time.time()
TOTAL_COST=0.0

class MediumTrainingCallback(tf.keras.callbacks.Callback):
    def __init__(self, lamb):
        global coarse_model
        super(MediumTrainingCallback, self).__init__()
        self.lamb = lamb
        self.submodel = coarse_model
        self.ds = train_COARSE
    
    def on_train_batch_end(self, batch, logs=None):
        global TOTAL_COST
        self.submodel.fit(
          self.ds,
          verbose=0,
          epochs=1,
          steps_per_epoch=self.lamb
        )
        TOTAL_COST += self.submodel.flopcost*self.lamb                

class CostCallback(tf.keras.callbacks.Callback):
    def __init__(self):
        super(CostCallback, self).__init__()
    
        
    def on_train_batch_end(self,batch,logs=None):
        global TOTAL_COST
        TOTAL_COST += self.model.flopcost

class LoggingAndSavingCallback(tf.keras.callbacks.Callback):
    def __init__(self):
        super(LoggingAndSavingCallback, self).__init__()
        self.best_weights=None
        self.best_validation=-1 
        
    def on_epoch_end(self, epoch, logs=None):
        global TOTAL_COST
        val = logs['val_loss']
        print(self.model.name + "_" + CYCLE, TOTAL_COST, val, time.time()-INITIAL_TIME)
        if self.best_validation < 0 or val < self.best_validation:
            self.best_validation = val           
            self.model.save_weights("saved_models/" + self.model.name + "_Pchange_"+ CYCLE + "/best_ver", save_format='tf')

class FineLoggingCallback(tf.keras.callbacks.Callback):
    def __init__(self):
        super(FineLoggingCallback, self).__init__()
        
    def on_epoch_end(self, epoch, logs=None):
        global TOTAL_COST, fine_model, normX, normY, test_idxs, BATCH_SIZE
        val1 = fine_prime.evaluate(normX[test_idxs],normY[test_idxs],batch_size=BATCH_SIZE,verbose=0)
        print(model.name+ "_"+ CYCLE, TOTAL_COST, val1[0], time.time()-INITIAL_TIME)
            
val1 = fine_prime.evaluate(normX[test_idxs],normY[test_idxs],batch_size=BATCH_SIZE,verbose=0)
print(model.name+ "_"+ CYCLE, 1.0, val1[0], 0.0)

def fine_train():
    fine_prime.fit(
              train_FINE,
              validation_data=valid_FINE,
              validation_freq=1,
              epochs=1,
              steps_per_epoch=20,
              verbose=0,          
              callbacks=[LoggingAndSavingCallback(),CostCallback()]
              )

def med_train():
    med_prime.fit(
              train_MED,
              validation_data=valid_MED,
              validation_freq=1,
              epochs=1,
              steps_per_epoch=20,
              verbose=0,
              callbacks=[CostCallback()]
              )

def coarse_train():
    coarse_prime.fit(
              train_COARSE,
              validation_data=valid_COARSE,
              validation_freq=1,
              epochs=1,
              steps_per_epoch=20,
              verbose=0,
              callbacks=[CostCallback()]
              )              

def recurTrain(lam, tmethods):
    if len(tmethods) == 0:
        return
    else:
        tmethods[0]()
        for i in range(lam):
            recurTrain(lam, tmethods[1:])
        tmethods[0]()
            
        
for i in range(1000):
    
    if CYCLE == 'V-CYCLE':
        recurTrain(1, [fine_train, med_train, coarse_train])
    
    elif CYCLE == 'W-CYCLE':
        recurTrain(2, [fine_train, med_train, coarse_train])
        
    elif CYCLE == 'GAMMA-3':
        recurTrain(3, [fine_train, med_train, coarse_train])
    
    else:
        
        fine_train()
    """

              

                       
    TOTAL_COST += 8*coarse_model.flopcost + 4*med_model.flopcost
    """



print(fine_model.evaluate(valid_FINE))

quit()




print(fine_model.evaluate(valid_FINE))

quit()

rando = np.random.random(size=(8,624,13))

print(medmodel(rando)[:4,:4,:])

medmodel.layers[4].weights = coarsemodel.layers[4].weights
#coarse_coarse_gcn.save_weights("temp.tf")
#med_coarse_gcn.load_weights("temp.tf")

print(medmodel(rando)[:4,:4,:])

quit()

optt = tf.keras.optimizers.Adam()
lossed = tf.keras.losses.MeanSquaredError()

model.compile(optimizer=optt,
              loss=lossed,
              metrics=['mse'])
model.summary()

unit_models = sorted([layer for layer in model.layers if "gcn_model" in layer.name], key=lambda x: x.name)
#[item.summary() for item in unit_models]

unit_models[0].trainable=False


model.compile(optimizer=optt,
              loss=lossed,
              metrics=['mse'])
model.summary()
quit()

INITIAL_TIME = time()

class LoggingAndSavingCallback(tf.keras.callbacks.Callback):
    def __init__(self):
        super(LoggingAndSavingCallback, self).__init__()
        self.best_weights=None
        self.best_validation=-1
        
    def on_epoch_end(self, epoch, logs=None):
        val = logs['val_loss']
        print(self.model.name, (epoch+1)*self.model.flopcost, val, time()-INITIAL_TIME)
        if self.best_validation < 0 or val < self.best_validation:
            self.best_validation = val           
            self.model.save_weights("saved_models/" + self.model.name + "/best_ver", save_format='tf')   

val1 = model.evaluate(normX[test_idxs],normY[test_idxs],batch_size=BATCH_SIZE,verbose=0)
print(model.name, 1.0, val1[0], 0.0)

model.fit(
          train_ds,
          validation_data=valid_ds,
          validation_freq=1,
          epochs=1000,
          steps_per_epoch=20,
          verbose=0,
          callbacks=[LoggingAndSavingCallback()]
          )
          
quit()          

@tf.function
def train_step(images, labels,i=0):
    with tf.GradientTape() as tape:
        predictions = model(images)
        loss = loss_object(labels, predictions)
        un_loss = loss_object((labels*stdY)+meanY, (predictions*stdY) + meanY) * 6.241
    gradients = tape.gradient(loss, model.trainable_variables)
    optimizer.apply_gradients(zip(gradients, model.trainable_variables))

    train_loss(loss)
    train_un_loss(un_loss)

loss_object = tf.keras.losses.MeanSquaredError()
train_loss = tf.keras.metrics.Mean(name='train_loss')
train_un_loss = tf.keras.metrics.Mean(name='train_un_loss')

optimizer = tf.keras.optimizers.Adam()

EPOCHS = 2

for epoch in range(EPOCHS):
    for i, (images, labels) in enumerate(train_ds):
        train_step(images, labels)

    #for test_images, test_labels in test_ds:
    #    test_step(test_images, test_labels)

    template = 'Epoch {}, Loss: {}, Un-Normalized Loss: {} x meV'
    print(template.format(epoch+1,
                        train_loss.result(), train_un_loss.result()))

    # Reset the metrics for the next epoch
    train_loss.reset_states()
    train_un_loss.reset_states()




